@@ -175,23 +175,32 @@ struct fy_parser {
 	struct fy_reader *reader;		/* the external reader, or ptr to builtin_reader */
 
 	struct fy_version default_version;
-	bool suppress_recycling : 1;
-	bool stream_start_produced : 1;
-	bool stream_end_produced : 1;
-	bool stream_end_reached : 1;
-	bool simple_key_allowed : 1;
-	bool tab_used_for_ws : 1;
-	bool stream_error : 1;
-	bool generated_block_map : 1;
-	bool last_was_comma : 1;
-	bool document_has_content : 1;
-	bool document_first_content_token : 1;
-	bool bare_document_only : 1;		/* no document start indicators allowed, no directives */
-	bool stream_has_content : 1;
-	bool parse_flow_only : 1;	/* document is in flow form, and stop parsing at the end */
-	bool parse_block_only : 1;	/* document is in embedded block form, and stop parsing at the end */
-	bool colon_follows_colon : 1;	/* "foo"::bar -> "foo": :bar */
-	bool had_directives : 1;	/* document had directives */
+	struct fy_version stream_version;	/* last version set in stream */
+
+	union {
+		struct {
+			bool suppress_recycling : 1;
+			bool stream_start_produced : 1;
+			bool stream_end_produced : 1;
+			bool stream_end_reached : 1;
+			bool simple_key_allowed : 1;
+			bool tab_used_for_ws : 1;
+			bool stream_error : 1;
+			bool generated_block_map : 1;
+			bool last_was_comma : 1;
+			bool document_has_content : 1;
+			bool document_first_content_token : 1;
+			bool bare_document_only : 1;		/* no document start indicators allowed, no directives */
+			bool stream_has_content : 1;
+			bool parse_flow_only : 1;	/* document is in flow form, and stop parsing at the end */
+			bool parse_block_only : 1;	/* document is in embedded block form, and stop parsing at the end */
+			bool colon_follows_colon : 1;	/* "foo"::bar -> "foo": :bar */
+			bool had_directives : 1;	/* document had directives */
+			bool is_checkpoint : 1;		/* a parser checkpoint (not a real parser) */
+		};
+		unsigned int parser_bitflags;
+	};
+
 	int flow_level;
 	int pending_complex_key_column;
 	struct fy_mark pending_complex_key_mark;
@@ -281,6 +290,7 @@ struct fy_parser {
 		struct fy_eventp_list args;
 		struct fy_document *fyd;
 	} mks;
+	struct fy_eventp *fyep_peek;
 };
 
 static inline struct fy_input *
@@ -331,6 +341,12 @@ static inline enum fy_flow_ws_mode fyp_fws_mode(const struct fy_parser *fyp)
 	return fy_reader_flow_ws_mode(fyp->reader);
 }
 
+static inline bool fyp_directive0_mode(const struct fy_parser *fyp)
+{
+	assert(fyp);
+	return fy_reader_directive0_mode(fyp->reader);
+}
+
 static inline bool fyp_block_mode(struct fy_parser *fyp)
 {
 	return !fyp_json_mode(fyp) && !fyp->flow_level;
@@ -390,13 +406,6 @@ static inline bool fyp_is_flow_blankz(const struct fy_parser *fyp, int c)
 	return fy_reader_is_flow_blankz(fyp->reader, c);
 }
 
-static inline const void *
-fy_ptr_slow_path(struct fy_parser *fyp, size_t *leftp)
-{
-	assert(fyp);
-	return fy_reader_ptr_slow_path(fyp->reader, leftp);
-}
-
 static inline const void *
 fy_ensure_lookahead_slow_path(struct fy_parser *fyp, size_t size, size_t *leftp)
 {
@@ -442,6 +451,13 @@ fy_parse_strncmp(struct fy_parser *fyp, const char *str, size_t n)
 	return fy_reader_strncmp(fyp->reader, str, n);
 }
 
+static FY_ALWAYS_INLINE inline int
+fy_parse_peek_at_offset_width(struct fy_parser *fyp, size_t offset, int *wp)
+{
+	assert(fyp);
+	return fy_reader_peek_at_offset_width(fyp->reader, offset, wp);
+}
+
 static FY_ALWAYS_INLINE inline int
 fy_parse_peek_at_offset(struct fy_parser *fyp, size_t offset)
 {
@@ -449,6 +465,13 @@ fy_parse_peek_at_offset(struct fy_parser *fyp, size_t offset)
 	return fy_reader_peek_at_offset(fyp->reader, offset);
 }
 
+static FY_ALWAYS_INLINE inline int
+fy_parse_peek_at_width_internal(struct fy_parser *fyp, int pos, ssize_t *offsetp, int *wp)
+{
+	assert(fyp);
+	return fy_reader_peek_at_width_internal(fyp->reader, pos, offsetp, wp);
+}
+
 static FY_ALWAYS_INLINE inline int
 fy_parse_peek_at_internal(struct fy_parser *fyp, int pos, ssize_t *offsetp)
 {
@@ -477,11 +500,25 @@ fy_is_generic_blankz_at_offset(struct fy_parser *fyp, size_t offset)
 	return fy_reader_is_generic_blankz(fyp->reader, fy_reader_peek_at_offset(fyp->reader, offset));
 }
 
+static FY_ALWAYS_INLINE inline int
+fy_parse_peek_at_width(struct fy_parser *fyp, int pos, int *wp)
+{
+	assert(fyp);
+	return fy_reader_peek_at_width(fyp->reader, pos, wp);
+}
+
 static FY_ALWAYS_INLINE inline int
 fy_parse_peek_at(struct fy_parser *fyp, int pos)
 {
 	assert(fyp);
-	return fy_reader_peek_at_internal(fyp->reader, pos, NULL);
+	return fy_reader_peek_at(fyp->reader, pos);
+}
+
+static FY_ALWAYS_INLINE inline int
+fy_parse_peek_width(struct fy_parser *fyp, int *wp)
+{
+	assert(fyp);
+	return fy_reader_peek_width(fyp->reader, wp);
 }
 
 static FY_ALWAYS_INLINE inline int
@@ -640,7 +677,7 @@ void fy_free_default(void *userdata, void *ptr);
 void *fy_realloc_default(void *userdata, void *ptr, size_t size);
 
 int fy_reader_fetch_flow_scalar_handle(struct fy_reader *fyr, int c, int indent, struct fy_atom *handle, bool sloppy_indent);
-int fy_reader_fetch_plain_scalar_handle(struct fy_reader *fyr, int c, int indent, int flow_level, struct fy_atom *handle, bool directive0);
+int fy_reader_fetch_plain_scalar_handle(struct fy_reader *fyr, int c, int indent, int flow_level, struct fy_atom *handle);
 
 void fy_reader_skip_ws_cr_nl(struct fy_reader *fyr);
 
@@ -662,4 +699,171 @@ void fy_parse_streaming_aliases_reset(struct fy_parser *fyp);
 struct fy_eventp *fy_parser_parse_resolve_prolog(struct fy_parser *fyp);
 struct fy_eventp *fy_parser_parse_resolve_epilog(struct fy_parser *fyp, struct fy_eventp *fyep);
 
+struct fy_parser_checkpoint {
+	struct fy_parser fyp_checkpoint;
+	int queued_token_count;
+	struct fy_token **queued_tokens;
+	int queued_input_count;
+	struct fy_input **queued_inputs;
+	bool diag_on_error;
+};
+
+/* diagnostics */
+
+static inline bool
+fyp_debug_log_level_is_enabled(struct fy_parser *fyp, enum fy_error_module module)
+{
+	return fyp && fy_diag_log_level_is_enabled(fyp->diag, FYET_DEBUG, module);
+}
+
+int fy_parser_vdiag(struct fy_parser *fyp, unsigned int flags,
+		    const char *file, int line, const char *func,
+		    const char *fmt, va_list ap);
+
+int fy_parser_diag(struct fy_parser *fyp, unsigned int flags,
+		   const char *file, int line, const char *func,
+		   const char *fmt, ...)
+			__attribute__((format(printf, 6, 7)));
+
+void fy_parser_diag_vreport(struct fy_parser *fyp,
+			    const struct fy_diag_report_ctx *fydrc,
+			    const char *fmt, va_list ap);
+void fy_parser_diag_report(struct fy_parser *fyp,
+			   const struct fy_diag_report_ctx *fydrc,
+			   const char *fmt, ...)
+		__attribute__((format(printf, 3, 4)));
+
+#ifdef FY_DEVMODE
+
+#define fyp_debug(_fyp, _module, _fmt, ...) \
+	do { \
+		struct fy_parser *__fyp = (_fyp); \
+		enum fy_error_module __module = (_module); \
+		\
+		if (fyp_debug_log_level_is_enabled(__fyp, __module)) \
+			fy_parser_diag(__fyp, FYET_DEBUG | FYDF_MODULE(__module), \
+					__FILE__, __LINE__, __func__, \
+					(_fmt) , ## __VA_ARGS__); \
+	} while(0)
+#else
+
+#define fyp_debug(_fyp, _module, _fmt, ...) \
+	do { } while(0)
+
+#endif
+
+#define fyp_info(_fyp, _fmt, ...) \
+	fy_parser_diag((_fyp), FYET_INFO, __FILE__, __LINE__, __func__, \
+			(_fmt) , ## __VA_ARGS__)
+#define fyp_notice(_fyp, _fmt, ...) \
+	fy_parser_diag((_fyp), FYET_NOTICE, __FILE__, __LINE__, __func__, \
+			(_fmt) , ## __VA_ARGS__)
+#define fyp_warning(_fyp, _fmt, ...) \
+	fy_parser_diag((_fyp), FYET_WARNING, __FILE__, __LINE__, __func__, \
+			(_fmt) , ## __VA_ARGS__)
+#define fyp_error(_fyp, _fmt, ...) \
+	fy_parser_diag((_fyp), FYET_ERROR, __FILE__, __LINE__, __func__, \
+			(_fmt) , ## __VA_ARGS__)
+
+#define fyp_scan_debug(_fyp, _fmt, ...) \
+	fyp_debug((_fyp), FYEM_SCAN, (_fmt) , ## __VA_ARGS__)
+#define fyp_parse_debug(_fyp, _fmt, ...) \
+	fyp_debug((_fyp), FYEM_PARSE, (_fmt) , ## __VA_ARGS__)
+#define fyp_doc_debug(_fyp, _fmt, ...) \
+	fyp_debug((_fyp), FYEM_DOC, (_fmt) , ## __VA_ARGS__)
+
+#define fyp_error_check(_fyp, _cond, _label, _fmt, ...) \
+	do { \
+		if (!(_cond)) { \
+			fyp_error((_fyp), _fmt, ## __VA_ARGS__); \
+			goto _label ; \
+		} \
+	} while(0)
+
+#define _FYP_TOKEN_DIAG(_fyp, _fyt, _type, _module, _fmt, ...) \
+	do { \
+		struct fy_diag_report_ctx _drc; \
+		memset(&_drc, 0, sizeof(_drc)); \
+		_drc.type = (_type); \
+		_drc.module = (_module); \
+		_drc.fyt = (_fyt); \
+		fy_parser_diag_report((_fyp), &_drc, (_fmt) , ## __VA_ARGS__); \
+	} while(0)
+
+#define FYP_TOKEN_DIAG(_fyp, _fyt, _type, _module, _fmt, ...) \
+	_FYP_TOKEN_DIAG(_fyp, fy_token_ref(_fyt), _type, _module, _fmt, ## __VA_ARGS__)
+
+#define FYP_PARSE_DIAG(_fyp, _adv, _cnt, _type, _module, _fmt, ...) \
+	_FYP_TOKEN_DIAG(_fyp, \
+		fy_token_create(FYTT_INPUT_MARKER, \
+			fy_fill_atom_at((_fyp), (_adv), (_cnt), \
+			alloca(sizeof(struct fy_atom)))), \
+		_type, _module, _fmt, ## __VA_ARGS__)
+
+#define FYP_MARK_DIAG(_fyp, _sm, _em, _type, _module, _fmt, ...) \
+	_FYP_TOKEN_DIAG(_fyp, \
+		fy_token_create(FYTT_INPUT_MARKER, \
+			fy_fill_atom_mark(((_fyp)), (_sm), (_em), \
+				alloca(sizeof(struct fy_atom)))), \
+		_type, _module, _fmt, ## __VA_ARGS__)
+
+#define FYP_NODE_DIAG(_fyp, _fyn, _type, _module, _fmt, ...) \
+	_FYP_TOKEN_DIAG(_fyp, fy_node_token(_fyn), _type, _module, _fmt, ## __VA_ARGS__)
+
+#define FYP_TOKEN_ERROR(_fyp, _fyt, _module, _fmt, ...) \
+	FYP_TOKEN_DIAG(_fyp, _fyt, FYET_ERROR, _module, _fmt, ## __VA_ARGS__)
+
+#define FYP_PARSE_ERROR(_fyp, _adv, _cnt, _module, _fmt, ...) \
+	FYP_PARSE_DIAG(_fyp, _adv, _cnt, FYET_ERROR, _module, _fmt, ## __VA_ARGS__)
+
+#define FYP_MARK_ERROR(_fyp, _sm, _em, _module, _fmt, ...) \
+	FYP_MARK_DIAG(_fyp, _sm, _em, FYET_ERROR, _module, _fmt, ## __VA_ARGS__)
+
+#define FYP_NODE_ERROR(_fyp, _fyn, _module, _fmt, ...) \
+	FYP_NODE_DIAG(_fyp, _fyn, FYET_ERROR, _module, _fmt, ## __VA_ARGS__)
+
+#define FYP_TOKEN_ERROR_CHECK(_fyp, _fyt, _module, _cond, _label, _fmt, ...) \
+	do { \
+		if (!(_cond)) { \
+			FYP_TOKEN_ERROR(_fyp, _fyt, _module, _fmt, ## __VA_ARGS__); \
+			goto _label; \
+		} \
+	} while(0)
+
+#define FYP_PARSE_ERROR_CHECK(_fyp, _adv, _cnt, _module, _cond, _label, _fmt, ...) \
+	do { \
+		if (!(_cond)) { \
+			FYP_PARSE_ERROR(_fyp, _adv, _cnt, _module, _fmt, ## __VA_ARGS__); \
+			goto _label; \
+		} \
+	} while(0)
+
+#define FYP_MARK_ERROR_CHECK(_fyp, _sm, _em, _module, _cond, _label, _fmt, ...) \
+	do { \
+		if (!(_cond)) { \
+			FYP_MARK_ERROR(_fyp, _sm, _em, _module, _fmt, ## __VA_ARGS__); \
+			goto _label; \
+		} \
+	} while(0)
+
+#define FYP_NODE_ERROR_CHECK(_fyp, _fyn, _module, _cond, _label, _fmt, ...) \
+	do { \
+		if (!(_cond)) { \
+			FYP_NODE_ERROR(_fyp, _fyn, _module, _fmt, ## __VA_ARGS__); \
+			goto _label; \
+		} \
+	} while(0)
+
+#define FYP_TOKEN_WARNING(_fyp, _fyt, _module, _fmt, ...) \
+	FYP_TOKEN_DIAG(_fyp, _fyt, FYET_WARNING, _module, _fmt, ## __VA_ARGS__)
+
+#define FYP_PARSE_WARNING(_fyp, _adv, _cnt, _module, _fmt, ...) \
+	FYP_PARSE_DIAG(_fyp, _adv, _cnt, FYET_WARNING, _module, _fmt, ## __VA_ARGS__)
+
+#define FYP_MARK_WARNING(_fyp, _sm, _em, _module, _fmt, ...) \
+	FYP_MARK_DIAG(_fyp, _sm, _em, FYET_WARNING, _module, _fmt, ## __VA_ARGS__)
+
+#define FYP_NODE_WARNING(_fyp, _fyn, _type, _module, _fmt, ...) \
+	FYP_NODE_DIAG(_fyp, _fyn, FYET_WARNING, _module, _fmt, ## __VA_ARGS__)
+
 #endif
